
#ifndef _PFFTSoundChannel_h_
#define _PFFTSoundChannel_h_


#define FFT_POINTS	256			// must be power of 2
#define MIN_DTMF_DURATION 50	// ms (for Australia at least)
// Number of consequetive FFT's that decode to the same DTMF before accepting decoding
#define MIN_DTMF_COUNT	((MIN_DTMF_DURATION/((FFT_POINTS*1000)/8000))+1)



class PFFTSoundChannel : public PSoundChannel
{
public:
	PFFTSoundChannel();
    PFFTSoundChannel(
      const PString & device,       /// Name of sound driver/device
      Directions dir,               /// Sound I/O direction
      unsigned numChannels = 1,     /// Number of channels eg mono/stereo
      unsigned sampleRate = 8000,   /// Samples per second
      unsigned bitsPerSample = 16   /// Number of bits per sample
	  );


    virtual BOOL Write(
      const void * buf, /// Pointer to a block of memory to write.
      PINDEX len        /// Number of bytes to write.
    );


private:
	void CommonConstruct(void);

	float m_realIn[FFT_POINTS];
	float m_realOut[FFT_POINTS];
	float m_imagOut[FFT_POINTS];
	float m_ampl[FFT_POINTS];
	short m_DTMFCount;
	char m_lastDTMF;
	char m_lastAcceptedDTMF;
	int m_lastOffset;



	static float FindDominantFrequency (
		int numSamples,           // total number of frequency samples in buffer
		int start,
		int stop,				 // number of positive frequencies to stop at
		float *real,             // real coefficient buffer
		float *imag,             // imaginary coefficient buffer
		float *ampl);             // work buffer for holding r*r+i*i values

	static unsigned NumberOfBitsNeeded ( unsigned PowerOfTwo );
	static unsigned ReverseBits ( unsigned index, unsigned NumBits );
	static void fft_float (
		unsigned  NumSamples,
		int       InverseTransform,
		float    *RealIn,
		float    *ImagIn,
		float    *RealOut,
		float    *ImagOut );

	static char FindDTMF(short freql, short freqh);

};


#endif
