/*
 * main.cxx
 *
 * PWLib application source file for OhPhone
 *
 * A H.323 "net telephone" application.
 *
 * Copyright (c) 1998-2000 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Open H323 Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Portions of this code were written with the assisance of funding from
 * Vovida Networks, Inc. http://www.vovida.com.
 *
 * Contributor(s): ______________________________________.
 *
 * $Log: main.cxx,v $
 * Revision 1.172  2000/07/14 12:46:10  craigs
 * Fixed problem with disabling tunnelling and changing sound volume
 *
 * Revision 1.171  2000/07/13 13:01:16  robertj
 * Split autoStartVideo so can select receive and transmit independently
 *
 * Revision 1.170  2000/07/10 23:45:01  dereks
 * Make reporting of play and record volume consistent.
 *
 * Revision 1.169  2000/07/10 23:35:28  dereks
 * Write error message if interactive changes to audio settings ignored.
 *
 * Revision 1.168  2000/07/10 02:50:45  dereks
 * Corrected connectport option.
 *
 * Revision 1.167  2000/07/09 07:09:04  robertj
 * Fixed strange line splits in strings. Also removed MSVC warning
 *
 * Revision 1.166  2000/07/08 19:50:15  craigs
 * Added options to set ports used for listening and connecting
 *
 * Revision 1.165  2000/07/04 12:58:26  craigs
 * Fixed problems with transmitted video size switching
 *
 * Revision 1.164  2000/07/04 06:47:39  craigs
 * Fixed video capability setup
 *
 * Revision 1.163  2000/07/04 05:11:23  robertj
 * Added error description if mixer open fails.
 *
 * Revision 1.162  2000/07/04 05:10:46  robertj
 * Win32 compatibility.
 *
 * Revision 1.161  2000/07/04 04:27:49  craigs
 * Fixed problem with autohook and normal sound card calls
 *
 * Revision 1.160  2000/07/03 10:33:37  craigs
 * Fixed autohook option to work with PhoneCARD
 *
 * Revision 1.159  2000/07/03 09:14:17  craigs
 * Seperated vide receive and transmit options
 *
 * Revision 1.158  2000/07/03 06:58:55  craigs
 * Fixed problems with videolocal parsing
 *
 * Revision 1.157  2000/07/03 06:35:19  craigs
 * Added ability for video to work on 24, 16 and 8 bit visuals
 *
 * Revision 1.156  2000/06/29 11:00:07  robertj
 * Added user interface for sound buffer depth adjustment.
 *
 * Revision 1.155  2000/06/20 02:38:32  robertj
 * Changed H323TransportAddress to default to IP.
 *
 * Revision 1.154  2000/06/19 00:32:26  robertj
 * Changed functionf or adding all lid capabilities to not assume it is to an endpoint.
 *
 * Revision 1.153  2000/06/12 02:36:08  craigs
 * Fixed problem with sound-recchan option
 * Added last number redial
 *
 * Revision 1.152  2000/06/10 11:10:55  rogerh
 * Enable LPC10 codec, but make it the lowest priority. Use -P LPC-10 to make it
 * the prefered codec if you want to try it.
 *
 * Revision 1.151  2000/06/10 10:59:39  rogerh
 * Update help to say that --h261 is used to Send Video as well as Receive it
 *
 * Revision 1.150  2000/06/10 10:31:02  rogerh
 * Fix ulaw and alaw entries in capName[] which prevented the G.711 ulaw and alaw
 * codecs from being Removed or Reordered in the codec preferences table.
 *
 * Revision 1.149  2000/06/07 05:50:22  robertj
 * Added call forwarding.
 *
 * Revision 1.148  2000/06/07 05:05:01  craigs
 * Fixed problem with reporting of round trip delays
 *
 * Revision 1.147  2000/06/05 04:45:15  robertj
 * Added LPC-10 2400bps codec
 *
 * Revision 1.146  2000/05/30 10:19:31  robertj
 * Added function to add capabilities given a LID.
 * Improved LID capabilities so cannot create one that is not explicitly supported.
 *
 * Revision 1.145  2000/05/29 01:41:33  craigs
 * Fixed problem with changing volume close to unity
 *
 * Revision 1.144  2000/05/25 12:06:19  robertj
 * Added PConfigArgs class so can save program arguments to config files.
 *
 * Revision 1.143  2000/05/25 08:44:16  craigs
 * Fixed some record/playback volume changes
 *
 * Revision 1.142  2000/05/25 02:41:18  robertj
 * Added missing #ifdef HAS_OSS all through the code.
 *
 * Revision 1.141  2000/05/24 14:22:24  craigs
 * Fixed problem with parsing quicknet volumes, and added volume display
 *
 * Revision 1.140  2000/05/24 12:27:45  craigs
 * Added new code to changeing of sound driver on the fly
 *
 * Revision 1.139  2000/05/23 12:57:40  robertj
 * Added ability to change IP Type Of Service code from applications.
 *
 * Revision 1.138  2000/05/23 11:32:40  robertj
 * Rewrite of capability table to combine 2 structures into one and move functionality into that class
 *    allowing some normalisation of usage across several applications.
 * Changed H323Connection so gets a copy of capabilities instead of using endponts, allows adjustments
 *    to be done depending on the remote client application.
 *
 * Revision 1.137  2000/05/22 14:28:07  craigs
 * Added time/date to output messages
 *
 * Revision 1.136  2000/05/18 02:27:31  craigs
 * Added ability to toggle silence supression in mid conversation
 * Fixed problem with caller ID reporting after new party format change
 *
 * Revision 1.135  2000/05/11 03:50:24  craigs
 * Improved ringFile functionality. Again!
 *
 * Revision 1.134  2000/05/11 02:05:22  craigs
 * Fixed problem with using ringfile
 *
 * Revision 1.133  2000/05/10 05:46:34  craigs
 * Another try at fixing the problem
 *
 * Revision 1.132  2000/05/10 05:38:46  craigs
 * Added ifdef guards for IXJ code and fixed compile error on Alpha
 *
 * Revision 1.131  2000/05/10 02:12:31  craigs
 * Added ability to play a file when ringing handset
 *
 * Revision 1.130  2000/05/10 00:02:36  robertj
 * Fixed missing #ifdef for setallocationbreakpoint command line argument.
 *
 * Revision 1.129  2000/05/09 11:20:52  craigs
 * Fixed statistics display and help message
 * Started adding mixer code
 *
 * Revision 1.128  2000/05/02 13:19:53  craigs
 * Removed "memory leaks"
 *
 * Revision 1.127  2000/05/02 04:32:25  robertj
 * Fixed copyright notice comment.
 *
 * Revision 1.126  2000/04/18 00:50:06  craigs
 * Display name of remote party when call is rejected
 *
 * Revision 1.125  2000/04/17 07:24:25  craigs
 * Fixed problem with timeouts and incorrect caller ID
 *
 * Revision 1.124  2000/04/11 04:52:24  craigs
 * Fixed problem with minimum and maximum stats being labelled incorrectly
 *
 * Revision 1.123  2000/04/11 03:58:44  robertj
 * Added new call end reasons for gatekeeper denied calls.
 *
 * Revision 1.122  2000/04/06 17:50:16  robertj
 * Added auto-start (including fast start) of video channels, selectable via boolean on the endpoint.
 *
 * Revision 1.121  2000/04/06 17:08:51  craigs
 * Fixed problems when compiling with X11 capable systems
 *
 * Revision 1.120  2000/04/05 18:25:00  robertj
 * Changed caller ID code for better portability.
 *
 * Revision 1.119  2000/04/05 16:28:27  craigs
 * Added caller ID functions
 *
 * Revision 1.118  2000/04/04 08:20:59  rogerh
 * Add 'null' h261 display option and fix broken x1132 display option
 *
 * Revision 1.117  2000/04/04 03:08:08  craigs
 * Fixed speakerphone mode, thanks to Ed Sternin for pointing this out
 *
 * Revision 1.116  2000/04/04 01:41:05  craigs
 * Improved error reporting when opening the IXJ device
 * Fixed problem with no busy tone after dialling unknown speed dial
 *
 * Revision 1.115  2000/04/03 23:53:46  robertj
 * Fixed warnings for when no IXJ card available.
 *
 * Revision 1.114  2000/03/29 21:02:58  robertj
 * Added function on LID to get available codecs.
 * Fixed display of duration in call end message.
 *
 * Revision 1.113  2000/03/29 02:18:50  robertj
 * Added stdout message with reason for call ending.
 *
 * Revision 1.112  2000/03/28 22:53:05  craigs
 * Added ability to change AEC in mid-call
 *
 * Revision 1.111  2000/03/25 02:04:45  robertj
 * Removed MSVC warning.
 *
 * Revision 1.110  2000/03/25 01:34:46  craigs
 * Changed name from voxilla to ohphone
 *
 * Revision 1.109  2000/03/20 20:21:32  craigs
 * Completely rewrote the state engine!
 *
 * Revision 1.108  2000/03/17 03:43:49  craigs
 * New state machine and extra features
 *
 * Revision 1.107  2000/03/04 20:00:34  craigs
 * Added ability to disable audio codecs
 *
 * Revision 1.106  2000/02/29 13:00:13  robertj
 * Added extra statistic display for RTP packets out of order.
 *
 * Revision 1.105  2000/02/25 13:39:50  craigs
 * Added ability to disable local video window via null video device
 *
 * Revision 1.104  2000/02/25 13:10:03  craigs
 * Added abioty to disable menu, and to dial after hangup
 *
 * Revision 1.103  2000/02/23 11:59:41  craigs
 * Added ability to set Quicknet record and playback volume, thanks to Fabio Ferrari
 *
 * Revision 1.102  2000/02/17 05:46:58  craigs
 * Removed redundant option processing, and added display of tunnelling disable option
 *
 * Revision 1.101  2000/02/17 05:33:55  craigs
 * Fixed error in help for tunnel disable.
 *
 * Revision 1.100  2000/02/17 05:28:55  craigs
 * Added H245 tunneling disable option
 *
 * Revision 1.99  2000/02/17 01:34:10  craigs
 * Fixed renaming if vinput option
 *
 * Revision 1.98  2000/02/16 05:34:35  craigs
 * Added help on new video parameters
 *
 * Revision 1.97  2000/02/10 03:08:02  craigs
 * Added ability to specify NTSC or PAL video format
 *
 * Revision 1.96  2000/02/04 19:47:52  craigs
 * Fixed problem with xlibvid.h only being included on Linux systems
 * Thanks to Roger Hardiman
 *
 * Revision 1.95  2000/02/04 05:17:31  craigs
 * New changes for video transmission from Derek Smithies
 *
 * Revision 1.94  2000/01/25 07:48:53  robertj
 * Added some missing removal of ixj code when HAS_IXJ is not defined.
 *
 * Revision 1.93  2000/01/13 04:11:58  robertj
 * Fixed non XLIB compile error.
 *
 * Revision 1.92  2000/01/13 04:03:46  robertj
 * Added video transmission
 *
 * Revision 1.91  2000/01/08 05:54:33  robertj
 * Added error message on opening sound device.
 *
 * Revision 1.90  2000/01/07 08:28:09  robertj
 * Additions and changes to line interface device base class.
 *
 * history deleted
 *
 * Revision 1.1  1998/12/14 09:13:19  robertj
 * Initial revision
 *
 */

#include <ptlib.h>

#include "main.h"
#include "gsmcodec.h"
#include "lpc10codec.h"
#include "h261codec.h"
#include "videoio.h"
#include "h323pdu.h"

#ifdef WIN32
#include "acmcodec.h"
#endif

#ifdef P_LINUX
#include "vidlinux.h"
#endif

#ifdef HAS_X11
#include "xlibvid.h"
#endif

#ifdef HAS_OSS
#define	DEFAULT_MIXER	"/dev/mixer"
#include	<linux/soundcard.h>
#endif

// uncomment below if to include xJack G729/G728 code
#ifdef _WIN32
#define	G729_G728
#endif

#include "version.h"
#include "PFFTSoundChannel.h"


PCREATE_PROCESS(OhPhone);

#define	DEFAULT_TIMEOUT	60000

class RingThread : public PThread
{
  PCLASSINFO(RingThread, PThread);

  public:
    RingThread(MyH323EndPoint & ep)
      : PThread(1000, NoAutoDeleteThread),
        endpoint(ep)
      { Resume(); }

    void Main()
      { endpoint.HandleRinging(); }

  protected:
    MyH323EndPoint & endpoint;
};

#define new PNEW


///////////////////////////////////////////////////////////////

OhPhone::OhPhone()
  : PProcess("Open H323 Project", "OhPhone",
             MAJOR_VERSION, MINOR_VERSION, BUILD_TYPE, BUILD_NUMBER)
{
}


OhPhone::~OhPhone()
{
}


void OhPhone::Main()
{
  //PArgList & args = GetArguments();
  PConfigArgs args(GetArguments());

  args.Parse(
             "a-auto-answer."        "-no-auto-answer."
             "b-bandwidth:"          "-no-bandwidth."
             "B-forward-busy:"       "-no-forward-busy."
             "d-autodial:"           "-no-autodial."
             "D-disable:"
             "e-silence."            "-no-silence."
             "f-fast-disable."       "-no-fast-disable."
             "F-forward-always:"     "-no-forward-always."
             "g-gatekeeper:"   
             "h-help."
             "i-interface:"          "-no-interface."
             "j-jitter:"             "-no-jitter."
             "l-listen."
             "n-no-gatekeeper."
             "N-forward-no-answer:"  "-no-forward-no-answer."
             "-answer-timeout:"      "-no-answer-timeout."
#if PTRACING
             "o-output:"             "-no-output."
#endif
             "p-proxy:"              "-no-proxy."
			 "-listenport:"			 "-no-listenport."
			 "-connectport:"		 "-no-connectport."
			 "-port:"		 		 "-no-port."
             "P-prefer:"
#ifdef HAS_IXJ
             "q-quicknet:"           "-no-quicknet."
#endif
             "r-require-gatekeeper." "-no-require-gatekeeper."
             "-save."
             "s-sound:"              "-no-sound."
             "-sound-in:"            "-no-sound-in."
             "-sound-out:"           "-no-sound-out."
             "-sound-buffers:"       "-no-sound-buffers."
#ifdef HAS_OSS
             "-sound-mixer:"         "-no-sound-mixer."
             "-sound-recchan:"       "-no-sound-recchan."
             "-sound-recvol:"        "-no-sound-recvol."
             "-sound-playvol:"       "-no-sound-playvol."
#endif
#ifdef _DEBUG
	     "-setallocationbreakpoint:"
#endif
             "T-h245tunneldisable."  "-no-h245tunneldisable."
#if PTRACING
             "t-trace."              "-no-trace."
#endif
             "-tos:"                 "-no-tos."
             "u-user:"               "-no-user."
             "v-verbose:"            "-no-verbose."

#ifdef WIN32
             "-acm:"                 "-no-acm."
#endif

#ifdef HAS_IXJ
             "-aec:"                 "-no-aec."
             "-disable-menu."        "-no-disable-menu."
             "-dial-after-hangup."   "-no-dial-after-hangup."
             "-callerid."            "-no-callerid."
	     "-autohook."            "-no-autohook."
             "-quicknet-recvol:"     "-no-quicknet-recvol."
             "-quicknet-playvol:"    "-no-quicknet-playvol."
#endif

#ifdef	G729_G728
             "-g728."                "-no-g728."
             "-g729."                "-no-g729."
#endif
             "-gsm."                 "-no-gsm."
             "-g711-ulaw."           "-no-g711-ulaw."
             "-g711-alaw."           "-no-g711-alaw."
             "-g7231."               "-no-g7231."
             "-h261:"                "-no-h261."
#ifdef WIN32
             "-ms-g7231."            "-no-ms-g7231."
#endif
             "-playvol:"             "-no-playvol."
             "-recvol:"              "-no-recvol."
	     "-ringfile:"
	     "-ringdelay:"

             "-videotransmit."       "-no-videotransmit."
             "-videolocal."          "-no-videolocal."
             "-videosize:"           "-no-videosize."
             "-videoformat:"         "-no-videoformat."
             "-videoinput:"          "-no-videoinput."              

             "-videoreceive:"        "-no-videoreceive."
             "-videoquality:"        "-no-videoquality."

          , FALSE);

#ifdef _DEBUG
  if (args.HasOption("setallocationbreakpoint"))
    PMemoryHeap::SetAllocationBreakpoint(args.GetOptionString("setallocationbreakpoint").AsInteger());
#endif

  BOOL disableMenu = args.HasOption("disable-menu");

  int verbose = 255;
  if (args.HasOption('v'))
    verbose = args.GetOptionString('v').AsInteger();

  if (verbose >= 3)
    cout << GetName()
         << " Version " << GetVersion(TRUE)
         << " by " << GetManufacturer()
         << " on " << GetOSClass() << ' ' << GetOSName()
         << " (" << GetOSVersion() << '-' << GetOSHardware() << ")\n\n";

#if PTRACING
  PTrace::SetOptions(PTrace::Timestamp | PTrace::Thread | PTrace::Blocks);
  PTrace::SetLevel(args.GetOptionCount('t'));
  PTextFile * output = NULL;
  if (args.HasOption('o')) {
    /* This is a bit tricky. We allocate the file for tracing on the heap, and
       we never actually deallocate it. So it come up as a leak on exit if
       memory checking is on. We cannot deallocate it as during termination
       some PTRACE's may get executed and attempt to trace to a closed (and
       deleted!) file object which crashes. Not good.
     */
    output = new PTextFile;
    if (output->Open(args.GetOptionString('o'), PFile::WriteOnly))
      PTrace::SetStream(output);
    else {
      if (verbose >= 1)
        cout << "Warning: could not open trace output file \""
             << args.GetOptionString('o') << '"' << endl;
      delete output;
    }
  }
  PTRACE(1, GetName()
         << " Version " << GetVersion(TRUE)
         << " by " << GetManufacturer()
         << " on " << GetOSClass() << ' ' << GetOSName()
         << " (" << GetOSVersion() << '-' << GetOSHardware() << ')');
#endif

  if (args.HasOption('h') || (!args.HasOption('l') && args.GetCount() == 0)) {
    cout << "Usage : " << GetName() << " [options] -l\n"
            "      : " << GetName() << " [options] [-p host] hostname/alias\n"
	    "\n   where:  hostname/alias = Remote host/alias to call\n"
            "\nOptions:\n"
            "  -a --auto-answer        : Automatically answer incoming calls\n"
            "  -d --autodial host      : Autodial host if phone off hook\n"
            "  -h --help               : Display this help message.\n"
            "  -l --listen             : Only listen for incoming calls\n"
            "  -v --verbose n          : Set amount of information displayed (0=none)\n"
            "  --disable-menu          : Disable internal menu\n"
            "  --ringfile filename     : Set sound file for \"ring\" annunciation\n"
            "  --ringdelay seconds     : Set delay between playing above file\n"
            "  --save                  : Save parameters in configuration file.\n"

            "\nGatekeeper options:\n"
            "  -g --gatekeeper host    : Specify gatekeeper host.\n"
            "  -n --no-gatekeeper      : Disable gatekeeper discovery.\n"
            "  -r --require-gatekeeper : Exit if gatekeeper discovery fails.\n"
            "  -p --proxy host         : Proxy/Gateway hostname/ip address\n"

            "\nDivert options:\n"
            "  -F --forward-always party    : Forward to remote party.\n"
            "  -B --forward-busy party      : Forward to remote party if busy.\n"
            "  -N --forward-no-answer party : Forward to remote party if no answer.\n"
            "     --answer-timeout time     : Time in seconds till forward on no answer.\n"

            "\nProtocol options:\n"
            "  -i --interface ipaddr   : Select interface to bind to for incoming connections (default is all interfaces)\n"
            "  --listenport            : Port to listen on for incoming connections (default 1720)\n"
            "  --connectport port      : Port to connect to for outgoing connections (default 1720)\n"
            "  -b --bandwidth bps      : Limit bandwidth usage to bps bits/second\n"
            "  -f --fast-disable       : Disable fast start\n"
            "  -T --h245tunneldisable  : Disable H245 tunnelling.\n"
            "  -u --user name          : Set local alias name(s) (defaults to login name)\n"
            "  --tos n                 : Set IP Type of Service byte to n\n"

            "\nAudio options:\n"
            "  -e --silence            : Disable silence detection for GSM and software G.711\n"
            "  -j --jitter delay       : Set jitter buffer to delay milliseconds\n"
            "  --recvol n              : Set record volume\n"
            "  --playvol n             : Set play volume\n"

            "\nVideo transmit options:\n"
            "  --videotransmit         : Enable video transmission\n"
            "  --videolocal            : Enable local video window\n"
            "  --videosize size        : Sets size of transmitted video window\n"
            "                             size can be small (default) or large\n"
            "  --videoformat type      : Set capture video format\n"
            "                             can be auto (default) pal or ntsc\n"
            "  --videoinput num        : Select capture video input (default is 0)\n"

            "\nVideo receive options:\n"
            "  --videoreceive viddev   : Receive video to following device\n"
            "                          :      null     do nothing\n"
            "                          :      ppm      create sequence of PPM files\n"
#ifdef HAS_VGALIB
            "                          :      svga256  256 colour VGA (Linux only)\n"
            "                          :      svga     full colour VGA (Linux only)\n"
#endif
#ifdef HAS_X11
	    "                          :      x11   automatically pick best X11 mode\n"
	    "                          :      x1124 X11 using 24 bit colour\n"
	    "                          :      x1116 X11 using 16 bit colour\n"
	    "                          :      x118  X11 using 8 bit grey scale\n"
            "  --videoquality n        : Set received video quality hint - 0 <= n <= 31\n"

            "\nSound card options:\n"
            "  -s --sound device       : Select sound card input/output device\n"
            "  --sound-in device       : Select sound card input device (overrides --sound)\n"
            "  --sound-out device      : Select sound card output device (overrides --sound)\n"
            "  --sound-buffers n       : Set sound buffer depth (default=2)\n"
#ifdef HAS_OSS
            "  --sound-mixer device    : Select sound mixer device (default is " DEFAULT_MIXER ")\n"
            "  --sound-recchan device  : Select sound mixer channel (default is mic)\n"
            "  --sound-recvol n        : Set record volume for sound card only (overrides --recvol)\n"
            "  --sound-playvol n       : Set play volume for sound card only (overrides --playvol)\n"
#endif

#ifdef HAS_IXJ
            "\nQuicknet card options:\n"
            "  -q -quicknet dev        : Use device (number or full device name)\n"
            "  --aec n                 : Set Audio Echo Cancellation level (0..3)\n"
            "  --autohook              : Don't use hook switch (for PhoneCard)\n"
            "  -c --callerid           : Enable caller id display\n"
            "  --dial-after-hangup     : Present dial tone after remote hang up\n"
            "  --quicknet-recvol n     : Set record volume for Quicknet card only (overrides recvol)\n"
            "  --quicknet-playvol n    : Set play volume for Quicknet card only (overrides playvol)\n"
#endif

            "\nAudio Codec options:\n"
            "  -D --disable codec      : Disable the specified codec (may be used multiple times)\n"
            "  -P --prefer codec       : Prefer the specified codec (may be used multiple times)\n"
#ifdef WIN32
            "  --acm num               : Use specified codec number from Windows ACM\n"
            "  --ms-g7231              : Use Microsoft ACM G.723.1 codec\n"
#endif
#ifdef HAS_IXJ
            "  --g7231                 : Set G.723.1 as preferred codec\n"
#endif
            "  --gsm                   : Set GSM 06.10 as preferred codec (default)\n"
            "  --g711-ulaw             : Set G.711 uLaw as preferred codec\n"
            "  --g711-alaw             : Set G.711 ALaw as preferred codec\n"
#ifdef	G729_G728
            "  --g728                  : Set G.728 as preferred codec\n"
            "  --g729                  : Set G.729 as preferred codec\n"
#endif
            "  --g7231                 : Set G.723.1 as preferred codec\n"
#endif
#if defined(PTRACING) || defined(_DEBUG)
            "\nDebug options:\n"
#endif
#if PTRACING
            "  -t --trace              : Enable trace, use multiple times for more detail\n"
            "  -o --output             : File for trace output, default is stderr\n"
#endif
#ifdef _DEBUG
	    "  --setallocationbreakpoint n : Enable breakpoint on memory allocation n\n"
#endif
            << endl;
    return;
  }

  args.Save("save");

  MyH323EndPoint * endpoint = new MyH323EndPoint;
  if (endpoint->Initialise(args, verbose)) {
    if (args.HasOption('l')) {
      if (verbose >= 2) 
        cout << "Waiting for incoming calls for \"" << endpoint->GetLocalUserName() << "\"\n";
    } else
      endpoint->MakeOutgoingCall(args[0], args.GetOptionString('p'));
    endpoint->AwaitTermination(!disableMenu);
  }

  delete endpoint;

  if (verbose >= 3)
    cout << GetName() << " ended." << endl;

#if PTRACING
  if (output != NULL)
    delete output;
#endif
}


///////////////////////////////////////////////////////////////

BOOL MyH323EndPoint::Initialise(PConfigArgs & args, int _verbose)
{
  PINDEX i;

  verbose = _verbose;
  uiState = uiDialtone;

  // get local username
  if (args.HasOption('u')) {
    PStringArray aliases = args.GetOptionString('u').Lines();
    SetLocalUserName(aliases[0]);
    for (i = 1; i < aliases.GetSize(); i++)
      AddAliasName(aliases[i]);
  }

#if 0
  PSoundChannel::writeDebug = TRUE;
  PSoundChannel::readDebug  = TRUE;
  OpalIxJDevice::writeDebug = TRUE;
  OpalIxJDevice::readDebug  = TRUE;
#endif


  terminateOnHangup    = !args.HasOption('l');
  silenceOn            = !args.HasOption('e');
  autoAnswer           = args.HasOption('a');
  noFastStart          = args.HasOption('f');
  doH245Tunnelling     = !args.HasOption('T');
  alwaysForwardParty   = args.GetOptionString('F');
  busyForwardParty     = args.GetOptionString('B');
  noAnswerForwardParty = args.GetOptionString('N');
  noAnswerTime         = args.GetOptionString("answer-timeout", "30").AsUnsigned();
  dialAfterHangup      = args.HasOption("dial-after-hangup");

  noAnswerTimer.SetNotifier(PCREATE_NOTIFIER(OnNoAnswerTimeout));

  if (args.HasOption("tos"))
    SetRtpIpTypeofService(args.GetOptionString("tos").AsUnsigned());

  if (args.HasOption('b')) {
    initialBandwidth = args.GetOptionString('b').AsUnsigned();
    if (initialBandwidth == 0) {
      cerr << "Illegal bandwidth specified." << endl;
      return FALSE;
    }
  }

  if (args.HasOption('j')) {
    unsigned jitter = args.GetOptionString('j').AsUnsigned();
    if (jitter >= 20 && jitter <= 10000)
      SetMaxAudioDelayJitter(jitter);
    else {
      cerr << "Jitter should be between 20 milliseconds and 10 seconds." << endl;
      return FALSE;
    }
  }

  if (args.HasOption("sound-buffers")) {
    soundChannelBuffers = args.GetOptionString("sound-buffers", "2").AsUnsigned();
    if (soundChannelBuffers < 2 || soundChannelBuffers > 99) {
      cerr << "Illegal sound buffers specified." << endl;
      return FALSE;
    }
  }

  if (verbose >= 3) {
    cout << "Local username: " << GetLocalUserName() << "\n"
         << "TerminateOnHangup is " << terminateOnHangup << "\n"
         << "Silence compression is " << silenceOn << "\n"
         << "Auto answer is " << autoAnswer << "\n"
         << "FastConnect is " << !noFastStart << "\n"
         << "H245Tunnelling is " << doH245Tunnelling << "\n"
         << "DialAfterHangup is " << doH245Tunnelling << "\n"
         << "Jitter buffer: "  << GetMaxAudioDelayJitter() << " ms\n"
         << endl;
  }

  if (args.HasOption("autodial")) {
    autoDial = args.GetOptionString("autodial");
    if (verbose >= 3)
      cout << "Autodial is set to "  << autoDial << "\n";
  }

  if (args.HasOption("h261")) {
    cout << "warning: --h261 option has been replaced by --videoreceive and --videotransmit" << endl;
    videoReceiveDevice = args.GetOptionString("h261");
  } else if (args.HasOption("videoreceive")) 
    videoReceiveDevice = args.GetOptionString("videoreceive");

  if (!videoReceiveDevice.IsEmpty()) {

    if (   !(videoReceiveDevice *= "ppm")
        && !(videoReceiveDevice *= "null")
#ifdef HAS_VGALIB
        && !(videoReceiveDevice *= "svga")
        && !(videoReceiveDevice *= "svga256")
#endif
#ifdef HAS_X11
        && !(videoReceiveDevice *= "x1132")
        && !(videoReceiveDevice *= "x1124")
        && !(videoReceiveDevice *= "x1116")
        && !(videoReceiveDevice *= "x118")
        && !(videoReceiveDevice *= "x11")
        && !(videoReceiveDevice *= "x1132s")
        && !(videoReceiveDevice *= "x1124s")
        && !(videoReceiveDevice *= "x1116s")
        && !(videoReceiveDevice *= "x118s")
        && !(videoReceiveDevice *= "x11s")
#endif
        ) {
        cerr << "Unknown video receive device \"" << videoReceiveDevice << "\"" << endl;
        return FALSE;
    }

    if (!args.HasOption("videoquality")) 
      videoQuality = -1;
    else {
      videoQuality = args.GetOptionString("videoquality").AsInteger();
      videoQuality = PMAX(0, PMIN(31, videoQuality));
    }
  }

  autoStartTransmitVideo = args.HasOption("videotransmit");
  if (autoStartTransmitVideo) {

    videoSize = 0; //Default is small.
    if (args.HasOption("videosize") && (args.GetOptionString("videosize") *= "large"))
      videoSize = 1;

    videoInput = 0; 
    if (args.HasOption("videoinput"))
      videoInput = args.GetOptionString("videoinput").AsInteger();

    videoIsPal = TRUE;
    if (args.HasOption("videoformat"))
      videoIsPal = args.GetOptionString("videoformat") *= "pal";

    videoLocal = args.HasOption("videolocal");
  }

  if (verbose >= 3) {
    if (videoReceiveDevice.IsEmpty())
      cout << "Video receive disabled" << endl << endl;
    else {
      cout << "Video receive using device : " << videoReceiveDevice << endl;
      cout << "Video receive quality hint : " << videoQuality << endl << endl;
    }
    if (!autoStartTransmitVideo)
      cout << "Video transmit disabled" << endl << endl;
    else {
      cout << "Video transmit enabled with local video window " << (videoLocal ? "en" : "dis") << "abled" << endl;
      cout << "Video transmit size is " << ((videoSize == 1) ? "large" : "small") << endl;
      cout << "Video capture using input " << videoInput << endl;
      cout << "Video capture using format " << (videoIsPal ? "PAL" : "NTSC") << endl << endl;
    }
  }

#ifdef HAS_IXJ
  if (args.HasOption('q')) {
    PString ixjDevice = args.GetOptionString('q');
    if (xJack.Open(ixjDevice)) {
      if (verbose >= 3)
        cout << "Using Quicknet " << xJack.GetName() << '\n';
      xJack.SetLineToLineDirect(0, 1, FALSE);
      xJack.EnableAudio(0, TRUE);

      callerIdEnable = args.HasOption("callerid");
      if (verbose >= 3)
        cout << "Caller ID set to " << callerIdEnable << endl;

      if (args.HasOption("aec")) {
        int aec = args.GetOptionString("aec").AsUnsigned();
        xJack.SetAEC(0, (OpalLineInterfaceDevice::AECLevels)aec);
        if (verbose >= 3)
          cout << "AEC set to " << aec << endl;
      } else {
        xJack.SetAEC(0, OpalLineInterfaceDevice::AECMedium);
        if (verbose >= 3)
          cout << "AEC set to default" << endl;
      }

      PString volStr;

      if (args.HasOption("quicknet-recvol"))
        volStr = args.GetOptionString("quicknet-recvol");
      else if (args.HasOption("recvol")) 
        volStr = args.GetOptionString("recvol");

      unsigned recvol;
      if (volStr.IsEmpty())
        xJack.GetRecordVolume(0, recvol);
      else {
        recvol = (unsigned)(volStr.AsReal() * 256.0);
        xJack.SetRecordVolume(0, recvol);
      }

      if (args.HasOption("quicknet-playvol"))
        volStr = args.GetOptionString("quicknet-playvol");
      else if (args.HasOption("playvol")) 
        volStr = args.GetOptionString("playvol");

      unsigned playvol;
      if (volStr.IsEmpty())
        xJack.GetPlayVolume(0, playvol);
      else {
        playvol = (unsigned)(volStr.AsReal() * 256.0);
        xJack.SetPlayVolume(0, playvol);
      }

      if (verbose >= 3) {
        cout << "Recording volume set to " << (recvol*1.0/256.0) << endl;
        cout << "Playing volume set to " << (playvol*1.0/256.0) << endl;
      }

      autoHook = args.HasOption("autohook");
      if (autoHook)
        xJack.StopTone(0);
      if (verbose >= 3)
        cout << "Autohook set to " << autoHook << endl;
    }
    else {
      cerr << "Could not open " << ixjDevice << ": ";
      int code = xJack.GetErrorNumber();
      if (code == EBADF)
        cerr << "check that the Quicknet driver is installed correctly" << endl;
      else if (code == EBUSY)
        cerr << "check that the Quicknet driver is not in use" << endl;
      else {
        PString errStr = xJack.GetErrorText();
        if (errStr.IsEmpty())
          errStr = psprintf("error code %i", code);
        cerr << errStr << endl;
      }
      return FALSE;
    }
  }

#endif

  if (args.HasOption("ringfile"))
    ringFile = args.GetOptionString("ringfile");
  if (args.HasOption("ringdelay"))
    ringDelay = args.GetOptionString("ringdelay").AsInteger();
  else
    ringDelay = 5;

  if (
#ifdef HAS_IXJ
     !xJack.IsOpen() ||
#endif
     args.HasOption("ringfile")) {
  
    if (!SetSoundDevice(args, "sound", PSoundChannel::Recorder))
      return FALSE;
    if (!SetSoundDevice(args, "sound", PSoundChannel::Player))
      return FALSE;
    if (!SetSoundDevice(args, "sound-in", PSoundChannel::Recorder))
      return FALSE;
    if (!SetSoundDevice(args, "sound-out", PSoundChannel::Player))
      return FALSE;

    if (verbose >= 3)
      cout << "Sound output device: \"" << GetSoundChannelPlayDevice() << "\"\n"
              "Sound  input device: \"" << GetSoundChannelRecordDevice() << "\"\n";

#ifdef HAS_OSS
    if (!InitialiseMixer(args, verbose))
      return FALSE;
#endif
  }

  // The order in which capabilities are added to the capability table
  // determines which one is selected by default.

#ifdef WIN32
  if (args.HasOption("acm")) {
    PStringArray acmNames = args.GetOptionString("acm").Lines();
    for (i = 0; i < acmNames.GetSize(); i++) {
      H323_ACMCapability * acm = new H323_ACMCapability(*this, acmNames[i].AsUnsigned());
      if (acm->IsValid())
        SetCapability(0, 0, acm);
      else {
        delete acm;
        cerr << "Invalid Windows ACM type (" << acmNames[i] << ") must be one of:\n";
        PStringArray names = H323_ACMCapability::GetCodecNames();
        for (i = 0; i < names.GetSize(); i++)
          cerr << "  " << names[i] << '\n';
        return FALSE;
      }
    }
  }
  if (args.HasOption("ms-g7231")) {
    H323_ACMG7231Capability * acm = new H323_ACMG7231Capability();
    if (acm->IsValid())
      SetCapability(0, 0, acm);
    else {
      delete acm;
      cerr << "Microsoft G.723.1 Codec is not available.\n";
      return FALSE;
    }
  }
#endif
#ifdef HAS_IXJ
  if (xJack.IsOpen())
    H323_LIDCapability::AddAllCapabilities(xJack, capabilities, 0, 0);
#endif

  SetCapability(0, 0, new H323_GSM0610Capability);
  SetCapability(0, 0, new H323_G711Capability(H323_G711Capability::muLaw));
  SetCapability(0, 0, new H323_G711Capability(H323_G711Capability::ALaw));
  SetCapability(0, 0, new H323_LPC10Capability(*this));

  PStringArray toRemove = args.GetOptionString('D').Lines();
  PStringArray toReorder = args.GetOptionString('P').Lines();

  static const char * const oldArgName[] = {
    "g7231",   "g729",  "g728",  "gsm", "g711-ulaw", "g711-alaw"
  };
  static const char * const capName[] = {
    "G.723.1", "G.729", "G.728", "GSM", "G.711 u-Law", "G.711 A-Law"
  };

  for (i = 0; i < PARRAYSIZE(oldArgName); i++) {
    if (args.HasOption(PString("no-")+oldArgName[i]))
      toRemove[toRemove.GetSize()] = capName[i];
    if (args.HasOption(oldArgName[i]))
      toReorder[toReorder.GetSize()] = capName[i];
  }

  capabilities.Remove(toRemove);
  capabilities.Reorder(toReorder);

  // if receiving video, add the capability and set it
  // if only trnsmitting, then add the capabilty but don't set it
  if (!videoReceiveDevice.IsEmpty()) {
    SetCapability(0, 1, new H323_H261Capability(2, 0, FALSE, FALSE, 6217));
    SetCapability(0, 1, new H323_H261Capability(0, 4, FALSE, FALSE, 6217));
  } else if (autoStartTransmitVideo) {
    AddCapability(new H323_H261Capability(2, 0, FALSE, FALSE, 6217));
    AddCapability(new H323_H261Capability(0, 4, FALSE, FALSE, 6217));
  }

  // if we want to transmit CIF images, reorder the capabilities
  if (videoSize == 1) {
    static const char * d[] = { " CIF", " QCIF" };
    PStringArray cifFirst(2, d);
    capabilities.Reorder(cifFirst);
  }

  SetCapability(0, P_MAX_INDEX, new H323_UserInputCapability);
  //SetCapability(0, P_MAX_INDEX, new H323_T120Capability);

  if (verbose >= 3)
    cout <<  "Codecs (in preference order):\n" << setprecision(2) << capabilities << endl << endl;

  H323ListenerTCP * listener;
  PIPSocket::Address interfaceAddress(INADDR_ANY);
  WORD port       = H323ListenerTCP::DefaultSignalPort;
  if (!args.GetOptionString("port").IsEmpty())
    args.GetOptionString("port").AsInteger();
  WORD listenPort = connectPort = port;

  if (!args.GetOptionString('i').IsEmpty())
    interfaceAddress = PIPSocket::Address(args.GetOptionString('i'));
  if (!args.GetOptionString("listenport").IsEmpty())
    listenPort = (WORD)args.GetOptionString("listenport").AsInteger();
  if (!args.GetOptionString("connectport").IsEmpty())
    connectPort = (WORD)args.GetOptionString("connectport").AsInteger();

  if (verbose >= 3) {
    cout << "Listening interface : " << interfaceAddress << endl;
    cout << "Listening port      : " << listenPort << endl;
    cout << "Connect port        : " << connectPort << endl;
  }
  listener  = new H323ListenerTCP(*this, interfaceAddress, listenPort);

  if (!StartListener(listener)) {
    cerr <<  "Could not open H.323 listener port on "
         << listener->GetListenerPort() << endl;
    return FALSE;
  }

  if (args.HasOption('g')) {
    PString gkName = args.GetOptionString('g');
    H323TransportUDP * rasChannel;
    if (args.GetOptionString('i').IsEmpty())
      rasChannel  = new H323TransportUDP(*this);
    else {
      PIPSocket::Address interfaceAddress(args.GetOptionString('i'));
      rasChannel  = new H323TransportUDP(*this, interfaceAddress);
    }
    if (SetGatekeeper(gkName, rasChannel)) {
      if (verbose >= 3) 
        cout << "Gatekeeper set: " << *gatekeeper << endl;
    } else {
      cerr << "Error registering with gatekeeper at \"" << gkName << '"' << endl;
      return FALSE;
    }
  }
  else if (!args.HasOption('n') || args.HasOption('r')) {
    if (verbose >= 2)
      cout << "Searching for gatekeeper..." << flush;
    if (DiscoverGatekeeper(new H323TransportUDP(*this))) {
      if (verbose >= 2) 
        cout << "\nGatekeeper found: " << *gatekeeper << endl;
    } else {
      if (verbose >= 2)
        cerr << "\nNo gatekeeper found." << endl;
      if (args.HasOption("require-gatekeeper")) 
        return FALSE;
    }
  }

  ringThread = NULL;

  return TRUE;
}

#ifdef HAS_OSS
BOOL MyH323EndPoint::InitialiseMixer(PConfigArgs & args, int _verbose)
{
  mixerDev = -1;

  // make sure mixer isn't disabled
  if (args.HasOption("no-sound-mixer"))
    return TRUE;

  PString mixerDeviceName = DEFAULT_MIXER;
  if (args.HasOption("mixer"))
    mixerDeviceName = args.GetOptionString("sound-mixer");

  mixerDev = ::open(mixerDeviceName, O_RDWR);
  if (mixerDev < 0) {
    cerr << "warning: Cannot open mixer device " << mixerDeviceName
         << ": " << ::strerror(errno) << endl;
    return TRUE;
  }

  char * mixerChanNames[] = SOUND_DEVICE_NAMES;
  int numMixerChans = SOUND_MIXER_NRDEVICES;

  // get the current record channel setting, and save it
  if (::ioctl(mixerDev, SOUND_MIXER_READ_RECSRC, &savedMixerRecChan) < 0) {
    cerr << "warning: cannot get current mixer record channels" << endl;
    savedMixerRecChan = -1;
  }

  // if the user specified a record channel, then find it
  // otherwise, find the currently select record channel
  if (args.HasOption("sound-recchan")) {
    PCaselessString mixerRecChanName = args.GetOptionString("sound-recchan");
    int i;
    for (i = 0; i < numMixerChans; i++) 
      if (mixerRecChanName *= mixerChanNames[i])
        break;
    if (i == numMixerChans) {
      cerr << "error: Cannot find record mixer channel " << mixerDeviceName << endl;
      return FALSE;
    }
    mixerRecChan = i;
  } else {
    int i;
    for (i = 0; i < numMixerChans; i++) 
      if (savedMixerRecChan & (1 << i))
        break;
    if (i == numMixerChans)
      mixerRecChan = SOUND_MIXER_MIC;
    else 
      mixerRecChan = i;
  } 

  PString volStr;
  if (args.HasOption("sound-recvol"))
    volStr = args.GetOptionString("sound-recvol");
  else if (args.HasOption("recvol")) 
    volStr = args.GetOptionString("recvol");

  if (volStr.IsEmpty()) {
    ::ioctl(mixerDev, MIXER_READ(mixerRecChan), &ossRecVol);
    ossRecVol &= 0xff;
  } else {
    ossRecVol = (unsigned)volStr.AsReal();
    int volVal = ossRecVol | (ossRecVol << 8);
    ::ioctl(mixerDev, MIXER_WRITE(mixerRecChan), &volVal);
  }

  if (args.HasOption("sound-playvol"))
    volStr = args.GetOptionString("sound-playvol");
  else if (args.HasOption("playvol")) 
    volStr = args.GetOptionString("playvol");

  if (volStr.IsEmpty()) {
    ::ioctl(mixerDev, SOUND_MIXER_READ_VOLUME, &ossPlayVol);
    ossPlayVol &= 0xff;
  } else {
    ossPlayVol = (unsigned)volStr.AsReal();
    int volVal = ossPlayVol | (ossPlayVol << 8);
    ::ioctl(mixerDev, SOUND_MIXER_WRITE_VOLUME, &volVal);
  }

  if (verbose >= 3) {
    cout << "Recording using mixer channel " << mixerChanNames[mixerRecChan] << endl;
    cout << "Record volume is " << ossRecVol << endl;
    cout << "Play volume is " << ossPlayVol << endl;
  }

  return TRUE;
}
#endif


BOOL MyH323EndPoint::SetSoundDevice(PConfigArgs & args,
                                    const char * optionName,
                                    PSoundChannel::Directions dir)
{
  if (!args.HasOption(optionName))
    return TRUE;

  PString dev = args.GetOptionString(optionName);

  if (dir == PSoundChannel::Player) {
    if (SetSoundChannelPlayDevice(dev))
      return TRUE;
  }
  else {
    if (SetSoundChannelRecordDevice(dev))
      return TRUE;
  }

  cerr << "Device for " << optionName << " (\"" << dev << "\") must be one of:\n";

  PStringArray names = PSoundChannel::GetDeviceNames(dir);
  for (PINDEX i = 0; i < names.GetSize(); i++)
    cerr << "  \"" << names[i] << "\"\n";

  return FALSE;
}


H323Connection * MyH323EndPoint::CreateConnection(unsigned callReference)
{
  return new MyH323Connection(*this, callReference, verbose);
}


BOOL MyH323EndPoint::OnIncomingCall(H323Connection & connection,
                                    const H323SignalPDU &,
                                    H323SignalPDU &)
{
  if (!alwaysForwardParty.IsEmpty()) {
    cout << "Forwarding call to \"" << alwaysForwardParty << "\"." << endl;
    return !connection.ForwardCall(alwaysForwardParty);
  }

  // incoming call is accepted if no call in progress 
  // unless the xJack is open and phone is off onhook

  if (currentCallToken.IsEmpty()
#ifdef HAS_IXJ
        && !(xJack.IsOpen() && (!autoHook && xJack.IsLineOffHook(OpalIxJDevice::POTSLine)))
#endif
  )
    return TRUE;

  if (busyForwardParty.IsEmpty()) {
    PTime now;
    cout << "Incoming call from \"" << connection.GetRemotePartyName() << "\" rejected at " << now << ", line busy!" << endl;
    return FALSE;
  }

  cout << "Forwarding call to \"" << busyForwardParty << "\"." << endl;
  return !connection.ForwardCall(busyForwardParty);
}


BOOL MyH323EndPoint::OnConnectionForwarded(H323Connection & /*connection*/,
                                           const PString & forwardParty,
                                           const H323SignalPDU & /*pdu*/)
{
  if (MakeCall(forwardParty, currentCallToken)) {
    cout << GetLocalUserName() << " is being forwarded to host " << forwardParty << endl;
    return TRUE;
  }

  cout << "Error forwarding call to \"" << forwardParty << '"' << endl;
  return FALSE;
}


void MyH323EndPoint::OnNoAnswerTimeout(PTimer &, INT)
{
  H323Connection * connection = FindConnectionWithLock(currentCallToken);
  if (connection != NULL) {
    cout << "Forwarding call to \"" << noAnswerForwardParty << "\"." << endl;
    connection->ForwardCall(noAnswerForwardParty);
    connection->Unlock();
  }
}


void MyH323EndPoint::OnConnectionEstablished(H323Connection & connection,
                                             const PString & token)
{
  currentCallToken = token;
  cout << "Call with \"" << connection.GetRemotePartyName() << "\" established." << endl;
  uiState = uiCallInProgress;
}


void MyH323EndPoint::OnConnectionCleared(H323Connection & connection, const PString & clearedCallToken)
{
  // stop any ringing that is occurring
  StopRinging();

  // ignore connections that are not the current connection
  if (clearedCallToken != currentCallToken)
    return;

  // indicate that our connection is now cleared
  currentCallToken = "";

  // indicate call has hungup
  uiState = uiCallHungup;

  if (verbose == 0)
    return;

  PString remoteName = '"' + connection.GetRemotePartyName() + '"';

  switch (connection.GetCallEndReason()) {
    case H323Connection::EndedByCallForwarded :
      return;   // Don't print message here, was printed when forwarded

    case H323Connection::EndedByRemoteUser :
      cout << remoteName << " has cleared the call";
      break;
    case H323Connection::EndedByCallerAbort :
      cout << remoteName << " has stopped calling";
      break;
    case H323Connection::EndedByRefusal :
      cout << remoteName << " did not accept your call";
      break;
    case H323Connection::EndedByNoAnswer :
      cout << remoteName << " did not answer your call";
      break;
    case H323Connection::EndedByTransportFail :
      cout << "Call with " << remoteName << " ended abnormally";
      break;
    case H323Connection::EndedByCapabilityExchange :
      cout << "Could not find common codec with " << remoteName;
      break;
    case H323Connection::EndedByNoAccept :
      cout << "Did not accept incoming call from " << remoteName;
      break;
    case H323Connection::EndedByAnswerDenied :
      cout << "Refused incoming call from " << remoteName;
      break;
    case H323Connection::EndedByNoUser :
      cout << "Gatekeeper could find user " << remoteName;
      break;
    case H323Connection::EndedByNoBandwidth :
      cout << "Call to " << remoteName << " aborted, insufficient bandwidth.";
      break;
    case H323Connection::EndedByConnectFail :
      switch (connection.GetSignallingChannel()->GetErrorNumber()) {
        case ENETUNREACH :
          cout << remoteName << " could not be reached.";
          break;
        case ETIMEDOUT :
          cout << remoteName << " is not online.";
          break;
        case ECONNREFUSED :
          cout << "No phone running for " << remoteName;
          break;
        default :
          cout << "Transport error calling " << remoteName;
      }
      break;
    default :
      cout << "Call with " << remoteName << " completed";
  }

  cout << ", duration "
       << setprecision(0) << setw(5)
       << (PTime() - connection.GetConnectionStartTime())
       << endl;
}


BOOL MyH323EndPoint::OpenAudioChannel(H323Connection & connection,
                                         BOOL isEncoding,
                                         unsigned bufferSize,
                                         H323AudioCodec & codec)
{
  codec.SetSilenceDetectionMode(silenceOn ? H323AudioCodec::AdaptiveSilenceDetection
                                          : H323AudioCodec::NoSilenceDetection);

#ifdef HAS_IXJ
  if (xJack.IsOpen()) {
    PTRACE(2, "xJack\tAttaching channel to codec");
    return codec.AttachChannel(new OpalLineChannel(xJack, 0, isEncoding));
  }
#endif

  if (isEncoding)
  {
	  if (H323EndPoint::OpenAudioChannel(connection, isEncoding, bufferSize, codec))
		return TRUE;
  }
  else	// only create a PFFTSoundChannel for received audio
  {
	  PString deviceName;
	  if (isEncoding)
		deviceName = GetSoundChannelRecordDevice();
	  else
		deviceName = GetSoundChannelPlayDevice();

	  PFFTSoundChannel * soundChannel = new PFFTSoundChannel;

	  if (soundChannel->Open(deviceName,
							 isEncoding ? PSoundChannel::Recorder
										: PSoundChannel::Player,
							 1, 8000, 16)) {
		PTRACE(3, "Codec\tOpened sound channel \"" << deviceName
			   << "\" for " << (isEncoding ? "record" : "play")
			   << "ing using " << soundChannelBuffers
			   << 'x' << bufferSize << " byte buffers.");
		soundChannel->SetBuffers(bufferSize, soundChannelBuffers);
		return codec.AttachChannel(soundChannel);
	  }

	  PTRACE(1, "Codec\tCould not open sound channel \"" << deviceName
			 << "\" for " << (isEncoding ? "record" : "play")
			 << "ing: " << soundChannel->GetErrorText());

	  delete soundChannel;
  }

  cerr << "Could not open sound device ";
  if (isEncoding)
    cerr << GetSoundChannelRecordDevice();
  else
    cerr << GetSoundChannelPlayDevice();
  cerr << " - Check permissions or full duplex capability." << endl;

  return FALSE;
}


BOOL MyH323EndPoint::OpenVideoDevice(H323Connection & connection,
                                     BOOL isEncoding,
                                     H323VideoCodec & codec)
{
  H323VideoDevice * channel = NULL;

  if (isEncoding) {

    cout << "Opening video encoder device" << endl;

    PAssert(autoStartTransmitVideo, "video encoder created without enable");

    codec.SetVideoInput(videoInput);
    codec.SetVideoFormat(videoIsPal);

#ifdef HAS_X11
    if (videoLocal)
      channel = new ShmXlibVideoDevice("Local");
    else
#endif
      channel = new NullVideoOutputDevice();

    //channel->SetSize(352>>(1-videoSize), 288>>(1-videoSize));
    channel->SetSize(codec.GetWidth(), codec.GetHeight());

  } else {

    PAssert(!videoReceiveDevice.IsEmpty(), "video decoder created without device type");

    PString nameStr = isEncoding ? PString("Local") : connection.GetRemotePartyName();

#ifdef HAS_X11
    // Dump received video to X11 window
    if (videoReceiveDevice.Left(3) *= "x11") {
      PString str = videoReceiveDevice;
      BOOL shared = str.Right(3) *= "s";
      str = str.Mid(3);
      if (!shared)
        channel = new XlibVideoDevice(nameStr);
      else {
        str = str.Left(str.GetLength()-1); 
        channel = new ShmXlibVideoDevice(nameStr);
      }
      int depth = str.AsInteger();
      if (depth > 0) 
        ((GenericXlibVideoDevice *)channel)->ForceDepth(depth);
    }
    else

#endif
#ifdef HAS_VGALIB
      // Dump received video to VGA 
    if (videoReceiveDevice *= "svga")
      channel = new LinuxSVGAFullOutputDevice();

    else if (videoReceiveDevice *= "svga256")
      channel = new LinuxSVGA256OutputDevice();

    else
#endif

    // Dump received video to PPM files
    if (videoReceiveDevice *= "ppm")
      channel = new PPMVideoOutputDevice(0);
    else

    // Dump received video to nowhere
    if (videoReceiveDevice *= "null")
      channel = new NullVideoOutputDevice();

    if (channel == NULL) { 
      PError << "unknown video output device \"" << videoReceiveDevice << "\"" << endl;
      return FALSE;
    }
  }

  PAssert(channel != NULL, "NULL video channel");

  return codec.AttachDevice(channel);
}


void MyH323EndPoint::MakeOutgoingCall(const PString & dest, const PString & proxy)
{
  PString fullAddress = dest;
  if ((fullAddress.Find(':') == P_MAX_INDEX) && (connectPort != H323ListenerTCP::DefaultSignalPort))
    fullAddress += psprintf(":%i", connectPort);

  if (!proxy)
    fullAddress += '@' + proxy;

  if (!MakeCall(fullAddress, currentCallToken)) {
    cout << "Error making call to \"" << dest << '"' << endl;
    return;
  }

  cout << GetLocalUserName() << " is calling host " << fullAddress << endl;
  uiState = uiConnectingCall;
}

void MyH323EndPoint::NewSpeedDial(const PString & ostr)
{
  PString str = ostr;
  PINDEX idx = str.Find(' ');
  if (str.IsEmpty() || (idx == P_MAX_INDEX)) {
    cout << "Must specify speedial number and string" << endl;
    return;
  }

  PString key  = str.Left(idx).Trim();
  PString data = str.Mid(idx).Trim();

  PConfig config("Speeddial");
  config.SetString(key, data);

  cout << "Speedial " << key << " set to " << data << endl;
}

void MyH323EndPoint::ListSpeedDials()
{
  PConfig config("Speeddial");
  PStringList keys = config.GetKeys();
  if (keys.GetSize() == 0) {
    cout << "No speed dials defined" << endl;
    return;
  }

  PINDEX i;
  for (i = 0; i < keys.GetSize(); i++) 
    cout << keys[i] << ":   " << config.GetString(keys[i]) << endl;
}

void MyH323EndPoint::StartCall(const PString & ostr)
{
  PString str = ostr;
  if (str.IsEmpty()) {
    cout << "Must supply hostname to connect to!\n";
    return;
  }

  else if ((str.GetLength() > 1) && (str[str.GetLength()-1] == '#')) {
    PString key = str.Left(str.GetLength()-1).Trim();
    PConfig config("Speeddial");
    str = config.GetString(key);
    if (str.IsEmpty()) {
      cout << "Speed dial \"" << key << "\" not defined" << endl;
      uiState = MyH323EndPoint::uiCallHungup;
    }
  }

  if (!str.IsEmpty()) {
    lastCalledParty = str;
    PINDEX idx = str.Find(' ');
    if (idx == P_MAX_INDEX)
      MakeOutgoingCall(str, PString());
    else {
      PString host = str.Left(idx).Trim();
      PString gw   = str.Mid(idx).Trim();
      MakeOutgoingCall(host, gw);
    }
  }
}

void MyH323EndPoint::AwaitTermination(BOOL hasMenu)
{
  PThread * userInterfaceThread = NULL;
  if (hasMenu)
    userInterfaceThread = new UserInterfaceThread(*this);

#ifndef HAS_IXJ

  exitFlag.Wait();

#else

  if (!xJack.IsOpen())
    exitFlag.Wait();
  else {

    speakerphoneSwitch = FALSE;
    BOOL oldOnHook     = TRUE;
    BOOL oldRealOnHook = TRUE;
    int  olduiState    = uiStateIllegal;
    PTime offHookTime;

    PString digits;

    // poll the handset every 100ms looking for state changes
    while (!exitFlag.Wait(100)) {

      // lock the user interface state whilst we change it
      uiStateMutex.Wait();

      // get real hook state
      BOOL realOnHook = !xJack.IsLineOffHook(OpalIxJDevice::POTSLine);
      BOOL onHook     = realOnHook;

      // if in speakerphone mode,
      if (speakerphoneSwitch) {

        // if the phone is onhook then don't look at the real hook state
        if (realOnHook)
          onHook = FALSE;

        // if the handset just went offhook, then get out of speakerphone mode
        else if (realOnHook != oldRealOnHook) {
          speakerphoneSwitch = FALSE;
          xJack.EnableAudio(0, TRUE);
          if (verbose > 1)
            cout << "Speakerphone off" << endl;
        }
      }

      // handle onhook/offhook transitions
      if (onHook != oldOnHook) {
        if (onHook) {
          digits = PString();
          HandleHandsetOnHook();
        } else {
          HandleHandsetOffHook();
          offHookTime = PTime();
        }
        HandleStateChange(onHook);
        olduiState = uiState;
      } 

      // handle timeouts and DTMF digits
      if (!onHook) {
        HandleHandsetTimeouts(offHookTime);
        HandleHandsetDTMF(digits);
      }

      // handle any state changes
      if (uiState != olduiState) {
        offHookTime = PTime();
        HandleStateChange(onHook);
      }

      // save hook state so we can detect changes
      oldOnHook     = onHook;
      oldRealOnHook = realOnHook;

      // save the old UI state so we can detect changes
      olduiState = uiState;

      uiStateMutex.Signal();
    }
  }
#endif
 
  if (userInterfaceThread != NULL) {
    userInterfaceThread->Terminate();
    userInterfaceThread->WaitForTermination();
    delete userInterfaceThread;
  }
}

#ifdef HAS_IXJ

#if 0
static char * stateNames[] = {
  "Dialtone",
  "AnsweringCall",
  "ConnectingCall",
  "WaitingForAnswer",
  "CallInProgress",
  "CallHungup",
  "StateIllegal"
};
#endif


void MyH323EndPoint::HandleStateChange(BOOL onHook)
{
  switch (uiState) {

    // dialtone whilst no call active
    case uiDialtone:
      if (onHook)
        xJack.RingLine(0, 0);
      else if (autoHook) {
        xJack.StopTone(0);
      } else {
        cout << "Playing dialtone" << endl;
        xJack.PlayTone(0, OpalLineInterfaceDevice::DialTone);
      }
      break;

    // no tone whilst waiting for remote party to connect
    case uiConnectingCall:
      if (onHook)
        xJack.RingLine(0, 0);
      else
        xJack.StopTone(0);
      break;

    // when connected, play ring tone
    case uiWaitingForAnswer:
      if (onHook)
        xJack.RingLine(0, 0);
      else {
        cout << "Playing ringtone" << endl;
        xJack.PlayTone(0, OpalLineInterfaceDevice::RingTone);
      }
      break;

    // when a call is in progress, stop all tones and remove DTMF tones from the stream
    case uiCallInProgress:
      if (onHook) 
        xJack.RingLine(0, 0);
      else {
        xJack.StopTone(0);
        xJack.SetRemoveDTMF(0, TRUE);
      }
      break;

    // remote end has hungup
    case uiCallHungup:
      if (terminateOnHangup)
        exitFlag.Signal();

      if (autoHook || onHook) {
        uiState = uiDialtone;
        xJack.RingLine(0, 0);
      } else {
        if (dialAfterHangup) 
          uiState = uiDialtone;
        else 
          xJack.PlayTone(OpalIxJDevice::POTSLine, OpalLineInterfaceDevice::BusyTone);
      }
      break;

    case uiAnsweringCall:
      if (autoHook || !onHook) 
        xJack.StopTone(0);
      else {
        xJack.SetCallerID(OpalIxJDevice::POTSLine, "");
        if (callerIdEnable) {
          H323Connection * connection = FindConnectionWithLock(currentCallToken);
          if (connection != NULL) {
            H323TransportAddress addr = connection->GetControlChannel().GetRemoteAddress();
            PIPSocket::Address ip;
            WORD port;
            addr.GetIpAndPort(ip, port);
            DWORD decimalIp = (ip[0] << 24) +
                              (ip[1] << 16) +
                              (ip[2] << 8) +
                               ip[3];
            PString remotePartyName = connection->GetRemotePartyName();
            PINDEX bracket = remotePartyName.Find('[');
            if (bracket != P_MAX_INDEX)
              remotePartyName = remotePartyName.Left(bracket);
            bracket = remotePartyName.Find('(');
            if (bracket != P_MAX_INDEX)
              remotePartyName = remotePartyName.Left(bracket);
            xJack.SetCallerID(OpalIxJDevice::POTSLine,
                              psprintf("%010li\t\t", decimalIp) + remotePartyName);
            connection->Unlock();
          }
        }
        xJack.RingLine(0, 0x33);
      }
      break;

   default:
      if (!onHook || autoHook) 
        xJack.StopTone(0);
      else
        xJack.RingLine(0, 0);
      break;
  }
}

void MyH323EndPoint::HandleHandsetOffHook()
{
  if (verbose > 1)
    cout << "Offhook - ";

//  if (speakerphoneSwitch) {
//    if (verbose > 1)
//      cout << "speakerphone off - ";
//    speakerphoneSwitch = FALSE;
//    xJack.EnableAudio(0, TRUE);
//  }
     
  switch (uiState) {

    case uiDialtone:
      if (!autoDial) {
        if (verbose > 1)
          cout << "auto-dialing " << autoDial << endl;
        StartCall(autoDial);
      } else {
        if (verbose > 1)
          cout << "dialtone" << endl;
      }
    break;
  
    case uiConnectingCall:
      if (verbose > 1)
        cout << "call connecting" << endl;
      break;   
  
    case uiAnsweringCall:
      if (verbose > 1)
        cout << "answering call" << endl;
      AnswerCall(H323Connection::AnswerCallNow);
      break;
  
    case uiWaitingForAnswer:
      if (verbose > 1)
        cout << "waiting for remote answer" << endl;
      break;

    case uiCallInProgress:  // should never occur!
      if (verbose > 1)
        cout << "call in progress" << endl;
      break;
  
    default:
      if (verbose > 1)
        cout << "not sure!" << endl;
      break;
  }
}

void MyH323EndPoint::HandleHandsetOnHook()
{
  if (uiState == uiCallHungup)
    uiState = uiDialtone;

  if (verbose > 1) 
    cout << "Onhook - ";

//#ifndef OLD_IXJ_DRIVER
//  if (speakerphoneSwitch) {
//    xJack.EnableAudio(0, FALSE);
//    if (verbose > 1)
//      cout << "speakerphone on." << endl;
//  } else
//#endif
//  {
    if (verbose > 1)
      cout << "ending call." << endl;
    ClearCall(currentCallToken);
//  }
  speakerphoneSwitch = FALSE;
}


void MyH323EndPoint::HandleHandsetDTMF(PString & digits)
{
  char newDigit = xJack.ReadDTMF(0);
  if (newDigit != '\0') {
    digits += newDigit;
    if (!digits) {
      switch (uiState) {
        case uiCallInProgress:
          {
            if (verbose > 1)
              cout << "Sending user indication message " << digits << endl;
            H323Connection * connection = FindConnectionWithLock(currentCallToken);
            if (connection != NULL) {
              connection->SendUserInput(digits);
              connection->Unlock();
            }
            digits = PString();
          }
          break;
    
        case uiDialtone:
          xJack.StopTone(0);
          if (digits.GetLength() > 0) {
            PINDEX i;
            for (i = 0; i < digits.GetLength(); i++)
              if (!isdigit(digits[i]))
                break;
            BOOL allDigits = i == digits.GetLength();
            if (digits[digits.GetLength()-1] == '#')  {
              if (!lastCalledParty.IsEmpty() && (digits.GetLength() == 1))
                StartCall(lastCalledParty);
              else if ((digits.GetLength() >= 9) && (digits[0] == '*')) {
                digits = digits.Mid(1, digits.GetLength()-2);
                digits.Replace('*', '.', TRUE);
                StartCall(digits);
              } else if (digits.GetLength() > 1) {
                StartCall(digits);
              }
              digits = PString();
            } else if (allDigits && (digits.GetLength() == 10)) {
              while (digits[0] == '0')
                digits = digits.Mid(1);
              if (digits.GetLength() > 0) {
                StartCall(digits);
                digits = PString();
              }
            }
          }
          break;
  
        default:
          break;
      }
    }
  }
}


void MyH323EndPoint::HandleHandsetTimeouts(const PTime & offHookTime)
{
  int timeout = 0;
  
  switch (uiState) {

    case uiDialtone:
      timeout = DEFAULT_TIMEOUT;
      break;

    case uiConnectingCall:
      timeout = DEFAULT_TIMEOUT;
      break;

    case uiWaitingForAnswer:
      timeout = DEFAULT_TIMEOUT;
      break;

    default:
      break;
  }

  if (timeout > 0) {
    PTime now;
    if ((now - offHookTime) > timeout) {
      if (verbose > 1)
        cout << "Operation timed out" << endl;
      ClearCall(currentCallToken);
      uiState = uiCallHungup;
    }
  }
}

#endif

void MyH323EndPoint::HandleUserInterface()
{
  PConsoleChannel console(PConsoleChannel::StandardInput);

  PTRACE(2, "OhPhone\tUser interface thread started.");

  PStringStream help;
  help << "Select:\n"
          "  0-9 : send user indication message\n"
          "  C   : connect to remote host\n"
          "  S   : Display statistics\n"
          "  H   : Hang up phone\n"
          "  L   : List speed dials\n"
          "  D   : Create new speed dial\n"
          "  {}  : Increase/reduce record volume\n"
          "  []  : Increase/reduce playback volume\n"
          "  V   : Display current volumes\n"
#ifdef HAS_IXJ
          "  A   : turn AEC up/down\n"
#endif
          "  E   : Turn silence supression on/off\n"
          "  F   : Forward call calls to address\n"
         ;

#ifdef HAS_IXJ
#ifndef OLD_IXJ_DRIVER
  if (xJack.IsOpen())
    help << "  P   : Enter speakerphone mode\n";
#endif
#endif

  help << "  X   : Exit program\n";

  for (;;) {

    // display the prompt
    cout << "Command ? " << flush;

    // terminate the menu loop if console finished
    char ch = (char)console.peek();
    if (console.eof()) {
      if (verbose)
        cerr << "\nConsole gone - menu disabled" << endl;
      return;
    }

    if (isdigit(ch)) {
      H323Connection * connection = FindConnectionWithLock(currentCallToken);
      if (connection == NULL) {
        cout << "No call in progress\n";
        console.ignore(INT_MAX, '\n');
      } else {
        PString str;
        console >> str;
        cout << "Sending user indication: " << str << endl;
        connection->SendUserInput(str);
        connection->Unlock();
      }
    }
    else {
      console >> ch;
      switch (tolower(ch)) {
        case '?' :
          cout << help << endl;
          break;

        case 'x' :
        case 'q' :
          cout << "Exiting." << endl;
          ClearAllCalls();
          uiState = uiDialtone;
          exitFlag.Signal();
          console.ignore(INT_MAX, '\n');
          return;

        case 'h' :
          if (!currentCallToken) {
            cout << "Hanging up call." << endl;
            if (!ClearCall(currentCallToken))
              cout << "Could not hang up current call!\n";
            speakerphoneSwitch = FALSE;
          }
          console.ignore(INT_MAX, '\n');
          break;

        case 'y' :
          AnswerCall(H323Connection::AnswerCallNow);
          console.ignore(INT_MAX, '\n');
          break;

        case 'n' :
          AnswerCall(H323Connection::AnswerCallDenied);
          console.ignore(INT_MAX, '\n');
          break;

        case 'c' :
          if (!currentCallToken.IsEmpty()) 
            cout << "Cannot make call whilst call in progress\n";
          else {
            PString str;
            console >> str;
            StartCall(str.Trim());
          }
          break;

        case 'l' :
          ListSpeedDials();
          break;

        case 'd' :
          {
            PString str;
            console >> str;
            NewSpeedDial(str.Trim());
          }
          break;

        case 'e' :
          if (currentCallToken.IsEmpty())
            cout << "No call in progress" << endl;
          else {
            H323Connection * connection = FindConnectionWithLock(currentCallToken);
            if (connection == NULL) 
              cout << "No connection active.\n";
            else {
              connection->Unlock();
              H323Channel * chan = connection->FindChannel(RTP_Session::DefaultAudioSessionID, FALSE);
              if (chan == NULL)
                cout << "Cannot find audio channel" << endl;
              else {
                H323Codec * rawCodec  = chan->GetCodec();
                if (!rawCodec->IsDescendant(H323AudioCodec::Class()))
                  cout << "Audio channel is not audio!" << endl;
                else {
                  H323AudioCodec * codec = (H323AudioCodec *)rawCodec;
                  H323AudioCodec::SilenceDetectionMode mode = codec->GetSilenceDetectionMode();
                  if (mode == H323AudioCodec::AdaptiveSilenceDetection) {
                    mode = H323AudioCodec::NoSilenceDetection;
                    cout << "Silence detection off" << endl;
                  } else {
                    mode = H323AudioCodec::AdaptiveSilenceDetection;
                    cout << "Silence detection on" << endl;
                  }
                  codec->SetSilenceDetectionMode(mode);
                }
              }
            //  connection->Unlock();
            }
          }
          break;

        case 's' :
          if (currentCallToken.IsEmpty())
            cout << "No call in progress" << endl;
          else {
            H323Connection * connection = FindConnectionWithLock(currentCallToken);
            if (connection == NULL) 
              cout << "No connection statistics available.\n";
            else {
              PTime now;
              PTime callStart = connection->GetConnectionStartTime();
              cout << "Connection statistics:\n   "
                   << "Remote party     : " << connection->GetRemotePartyName() << "\n   "
                   << "Start            : " << callStart << "\n   "
                   << "Duration         : " << setw(5) << setprecision(0) << (now - callStart) << " mins\n   "
                   << "Round trip delay : " << connection->GetRoundTripDelay().GetMilliSeconds() << " msec"
                   << endl;
              RTP_Session * session = connection->GetSession(RTP_Session::DefaultAudioSessionID);
              if (session == NULL)
                cout << "No RTP session statistics available.\n";
              else
                cout << "RTP session statistics:\n   "

                     << session->GetPacketsSent() << '/'
                     << session->GetOctetsSent() 
                     << " packets/bytes sent\n   "

                     << session->GetMaximumSendTime() << '/'
                     << session->GetAverageSendTime() << '/'
                     << session->GetMinimumSendTime() << " max/avg/min send time\n   "

                     << session->GetPacketsReceived() << '/'
                     << session->GetOctetsReceived()
                     << " packets/bytes received\n   "

                     << session->GetMaximumReceiveTime() << '/'
                     << session->GetAverageReceiveTime() << '/'
                     << session->GetMinimumReceiveTime() << " max/avg/min receive time\n   "

                     << session->GetPacketsLost() << '/'
                     << session->GetPacketsOutOfOrder() << '/'
                     << session->GetPacketsTooLate() 
                     << " packets dropped/out of order/late\n   "

		     << endl;


              connection->Unlock();
            }
          }
          break;

#ifdef HAS_IXJ
#ifndef OLD_IXJ_DRIVER
        case 'p' :
          if (xJack.IsOpen()) {
            speakerphoneSwitch = !speakerphoneSwitch;
            xJack.EnableAudio(0, !speakerphoneSwitch);
            if (verbose > 1)
              cout << "Speakerphone " 
                   << (speakerphoneSwitch ? "on" : "off")
                   << endl;
          }
          console.ignore(INT_MAX, '\n');
          break;
#endif

        case 'a' :
          if (xJack.IsOpen()) {
            int aec = xJack.GetAEC(0);
            if (ch == 'a')
              aec--;
            else
              aec++;
            if (aec < 0)
              aec = OpalLineInterfaceDevice::AECHigh;
            else if (aec > OpalLineInterfaceDevice::AECHigh)
              aec = OpalLineInterfaceDevice::AECOff;
            xJack.SetAEC(0, (OpalLineInterfaceDevice::AECLevels)aec);
            if (verbose > 2)
              cout << "New AEC level is " << aec << endl;
          } else
              cout <<"AEC change ignored as xJack closed"<<endl;
          break;
#endif
        case 'v' :
#ifdef HAS_IXJ
          if (xJack.IsOpen()) {
            unsigned vol;
            xJack.GetPlayVolume(0, vol);
            cout << "Play volume is " << (vol*1.0/256.0) << endl;
            xJack.GetRecordVolume(0, vol);
            cout << "Record volume is " << (vol*1.0/256.0) << endl;
          }
#endif
#if defined(HAS_IXJ) && defined(HAS_OSS)
          else
#endif
#ifdef HAS_OSS
          {
            cout << "Play volume is " << ossPlayVol << endl;
            cout << "Record volume is " << ossRecVol << endl;
          }
#endif
          break;

        case '[' :
        case ']' :
        case '{' :
        case '}' :
#ifdef HAS_IXJ
          if (xJack.IsOpen()) {
            unsigned vol;
            if (ch == '{' || ch == '}')
              xJack.GetRecordVolume(0, vol);
            else
              xJack.GetPlayVolume(0, vol);

            int step;
            if (ch == '[' || ch == '{') {
              if (vol <= 0x100)
                step = -52;
              else if (vol < (0x100 + 0x100))
                step = 0x100 - vol;
              else
                step = -0x100;
            } else {
              if (vol < (0x100 - 52))
                step = 52;
              else if (vol < 0x100)
                step = 0x100 - vol;
              else
                step = 0x100;
            }

            int newvol = vol + step;
            if (newvol > 0x1000) 
              newvol = 0x1000;
            else if (newvol < 0)
              newvol = 0x00;

            if (ch == '{' || ch == '}') {
              xJack.SetRecordVolume(0, (unsigned)newvol);
              if (verbose > 2)
               cout << "Record volume is " << (newvol*1.0/256.0) << endl;
            } else {
              xJack.SetPlayVolume(0, (unsigned)newvol);
              if (verbose > 2)
               cout << "Play volume is " << (newvol*1.0/256.0) << endl;
            }
          } 
#endif
#if defined(HAS_IXJ) && defined(HAS_OSS)
          else
#endif
#ifdef HAS_OSS
          {
            int vol;
            if (ch == '{' || ch == '}')
              vol = ossRecVol;
            else
              vol = ossPlayVol;
            
            vol += ((ch == '[') || (ch == '{')) ? -10 : 10;
            if (vol < 0)
              vol = 0;
            else if (vol > 100)
              vol = 100;

            if (mixerDev >= 0)  {
              int volVal = vol | (vol << 8);
              if (ch == '{' || ch == '}') {
                ossRecVol = vol;
                ::ioctl(mixerDev, MIXER_WRITE(mixerRecChan), &volVal);
                cout << "Record volume is " << ossRecVol << endl;
              } else {
                ossPlayVol = vol;
                ::ioctl(mixerDev, SOUND_MIXER_WRITE_VOLUME, &volVal);
                cout << "Play volume is " << ossPlayVol << endl;
              }
            } else
              cout << "Audio setting change ignored as mixer device disabled"<<endl;
          }
#endif
          break;

        case 'f' :
          console >> alwaysForwardParty;
          alwaysForwardParty = alwaysForwardParty.Trim();
          if (!alwaysForwardParty)
            cout << "Forwarding all calls to \"" << alwaysForwardParty << '"' << endl;
          else
            cout << "Call forwarding of all calls disabled." << endl;
          break;

        default:
          cout << "Unknown command " << ch << endl;
          console.ignore(INT_MAX, '\n');
          break;
      }
    }
  }
}


void MyH323EndPoint::AnswerCall(H323Connection::AnswerCallResponse response)
{
  if (uiState != uiAnsweringCall)
    return;

  StopRinging();

  H323Connection * connection = FindConnectionWithLock(currentCallToken);
  if (connection == NULL)
    return;

  connection->AnsweringCall(response);
  connection->Unlock();

  if (response == H323Connection::AnswerCallNow) {
    cout << "Accepting call." << endl;
    uiState = uiCallInProgress;
  } else {
    cout << "Rejecting call." << endl;
    uiState = uiCallHungup;
  }
}


void MyH323EndPoint::HandleRinging()
{
  PSoundChannel dev(GetSoundChannelPlayDevice(), PSoundChannel::Player);
  if (!dev.IsOpen()) {
    PTRACE(2, "Cannot open sound device for ring");
    return;
  }

  if (ringDelay < 0) {
    PTRACE(2, "Playing " << ringFile);
    dev.PlayFile(ringFile, TRUE);
  } else {
    PTimeInterval delay(0, ringDelay);
    PTRACE(2, "Playing " << ringFile << " with repeat of " << delay << " ms");
    do {
      dev.PlayFile(ringFile, TRUE);
    } while (!ringFlag.Wait(delay));
  }
}


void MyH323EndPoint::StartRinging()
{
  PAssert(ringThread == NULL, "Ringing thread already present");

  if (!noAnswerForwardParty)
    noAnswerTimer = PTimeInterval(0, noAnswerTime);

  if (!ringFile)
    ringThread = new RingThread(*this);
}


void MyH323EndPoint::StopRinging()
{
  noAnswerTimer.Stop();

  if (ringThread == NULL)
    return;

  ringFlag.Signal();
  ringThread->WaitForTermination();
  delete ringThread;
  ringThread = NULL;
}


///////////////////////////////////////////////////////////////

MyH323Connection::MyH323Connection(MyH323EndPoint & ep, unsigned callReference, int _verbose)
  : H323Connection(ep, callReference),
    myEndpoint(ep), verbose(_verbose)
{
  h245Tunneling = ep.doH245Tunnelling;
}


H323Connection::AnswerCallResponse
     MyH323Connection::OnAnswerCall(const PString & caller,
                                    const H323SignalPDU & /*setupPDU*/,
                                    H323SignalPDU & /*connectPDU*/)
{
  PTRACE(1, "H225\tOnWaitForAnswer");

  PTime now;

  if (myEndpoint.autoAnswer) {
    cout << "Automatically accepting call at " << now << endl;
    return AnswerCallNow;
  }

  myEndpoint.currentCallToken = GetCallToken();
  myEndpoint.uiState = MyH323EndPoint::uiAnsweringCall;


  cout << "Incoming call from \""
       << caller
       << "\" at "
       << now
       << ", answer call (Y/n)? "
       << flush;

  myEndpoint.StartRinging();
  
  return AnswerCallPending;
}


void MyH323Connection::OnSelectLogicalChannels()
{
  if (fastStartState != FastStartDisabled && myEndpoint.noFastStart)
    return;

#if 1 // Clear to temporarily disable the transmitter start up
  H323Connection::OnSelectLogicalChannels();
#endif
}


BOOL MyH323Connection::OnOpenLogicalChannel(const H245_OpenLogicalChannel & openPDU,
                                            H245_OpenLogicalChannelAck & ackPDU,
                                            unsigned & errorCode)
{
#if 0 // Set to temporarily disable the receiver acceptance
  return FALSE;
#else
  return H323Connection::OnOpenLogicalChannel(openPDU, ackPDU, errorCode);
#endif
}


BOOL MyH323Connection::OnStartLogicalChannel(H323Channel & channel)
{
  if (!H323Connection::OnStartLogicalChannel(channel))
    return FALSE;

  if (verbose >= 2) {
    cout << "Started logical channel: ";

    switch (channel.GetDirection()) {
      case H323Channel::IsTransmitter :
        cout << "sending ";
        break;

      case H323Channel::IsReceiver :
        cout << "receiving ";
        break;

      default :
        break;
    }

    cout << channel.GetCapability() << endl;  
  }

  if (channel.GetDirection() == H323Channel::IsReceiver) {
    int videoQuality = myEndpoint.videoQuality;

    if (channel.GetCodec()->IsDescendant(H323VideoCodec::Class()) && (videoQuality >= 0)) {

      //const H323_H261Capability & h261Capability = (const H323_H261Capability &)channel.GetCapability();
      //if (!h261Capability.GetTemporalSpatialTradeOffCapability()) 
      //  cout << "Remote endpoint does not allow video quality configuration" << endl;
      //else {
        cout << "Requesting remote endpoint to send video quality " << videoQuality << "/31" << endl;
  
        // kludge to wait for channel to ACK to be sent
        PThread::Current()->Sleep(2000);
  
        H323ControlPDU pdu;
        H245_CommandMessage & command = pdu.Build(H245_CommandMessage::e_miscellaneousCommand);
  
        H245_MiscellaneousCommand & miscCommand = command;
        miscCommand.m_logicalChannelNumber = (unsigned)channel.GetNumber();
        miscCommand.m_type.SetTag(H245_MiscellaneousCommand_type::e_videoTemporalSpatialTradeOff);
        PASN_Integer & value = miscCommand.m_type;
        value = videoQuality;
        WriteControlPDU(pdu);
      //}
    }
  }

  return TRUE;
}


BOOL MyH323Connection::OnAlerting(const H323SignalPDU & /*alertingPDU*/,
                                  const PString & username)
{
  PAssert((myEndpoint.uiState == MyH323EndPoint::uiConnectingCall) ||
          (myEndpoint.uiState == MyH323EndPoint::uiWaitingForAnswer),
     psprintf("Alerting received in state %i whilst not waiting for incoming call!", myEndpoint.uiState));

  if (verbose > 0)
    cout << "Ringing phone for \"" << username << "\" ..." << endl;

  myEndpoint.uiState = MyH323EndPoint::uiWaitingForAnswer;

  return TRUE;
}


void MyH323Connection::OnUserInputString(const PString & value)
{
  cout << "User input received: \"" << value << '"' << endl;
}


// End of File ///////////////////////////////////////////////////////////////
