/*
 * xlibvid.h
 *
 *This file contains the class heirachy for 1)creating windows under xlib.
 *                                          2)displaying data on those windows
 *                                          3)converting image formats.
 *
 * Copyright (c) 1999-2000 Indranet Technologies ltd
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Open H323
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Contributor(s): Author Derek J Smithies (derek@indranet.co.nz)
 *
 * Note the license of this file was changed to the above with the full
 * permission of the authors.
 *
 * $Log: xlibvid.h,v $
 * Revision 1.5  2000/07/03 06:35:20  craigs
 * Added ability for video to work on 24, 16 and 8 bit visuals
 *
 * Revision 1.4  2000/05/02 04:32:25  robertj
 * Fixed copyright notice comment.
 *
 */

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/extensions/XShm.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/shm.h>
#include <sys/types.h>
#include <sys/ipc.h>

#ifndef __XLIBVID_H
#define __XLIBVID_H

#include <ptlib.h>
#include <videoio.h>

class GenericXlibVideoDevice : public H323VideoDevice
{
  PCLASSINFO(GenericXlibVideoDevice, H323VideoDevice);

 public:
  // Constructor without window size arguements does not make a window
  GenericXlibVideoDevice(const PString & _remoteName)
    { remoteName = _remoteName; width=0; height=0; n_bytes=0; forceDepth = 0; }

  ~GenericXlibVideoDevice()
    { CloseWindow(); }

   virtual BOOL Redraw (const BYTE *frame);

   virtual void SetSize (int _width ,int _height);

   virtual void Close()
     {  CloseWindow(); }

   virtual PString GetRemoteName() const
     { return remoteName ; }
    
   virtual void SetRemoteName(const PString & _remoteName)
     { remoteName = _remoteName; }

   void ForceDepth(int d)
     { forceDepth = d; }

  protected:
    // overrides from H323VideoDevice
    virtual  BOOL WriteLineSegment(int x, int y, unsigned len, const BYTE * data)
       { return TRUE; }

    // new functions

    // translate the  411 input format to the correct format
    //virtual void Translate411Input(BYTE *dest, const BYTE *frame) = 0;
    void Translate411Input(BYTE *dest, const BYTE *frame);

    // allocate resources required
    virtual void OpenRender(Visual * xvisual) = 0;

    // dispose resources required
    virtual void CloseRender() = 0;

    // display the image
    virtual void DisplayImage(XImage *xi) = 0;

    BOOL OpenWindow(int _width = 0, int _height = 0);
    void CloseWindow();

    void MaskCalc(int & pos, int & wid, int & newMask, int mask);
    void GenerateImage(BYTE * dest);
    void FillRect(BYTE * dest, int x, int y, int width, int height, BYTE R, BYTE G, BYTE B);
    void SetPixel(BYTE * dest, int x, int y, BYTE r, BYTE g, BYTE b);

    // Translate from 411 format to xxx format stuff.
    void Translate411ToRGB32(BYTE *dest, const BYTE *frame);

    // member variables
    Window  win;
    Display *display;
    int     screen;
    GC      gc;
    Colormap colormap;
    PString remoteName;
    int     pixelDepth;
    int     byteDepth;
    int     forceDepth;

    int redWidth, greenWidth, blueWidth;
    int redPos, greenPos, bluePos;
    int redMask, greenMask, blueMask;
    BOOL useGreyScale;
    long greyColors[64];

    int     n_bytes; //# bytes in Y plane of image (i.e. width*height)
    XImage  *images[2];
    BYTE    *pixels[2];

    //    u_int  uvtab[65536]; //Gives quick conversion of YUV to RGB format.
    // However, causes too cache thrashing, and it is quicker to use arithmetic
    // to determine the appropriate RGB value, given a YUV value.
    u_int  omask;
    u_int  pmask;
    int    useImage0;
    BYTE   ri,gi,bi,ai;

    PMutex xLibMutex;
};

class ShmXlibVideoDevice : public GenericXlibVideoDevice
{
  PCLASSINFO(ShmXlibVideoDevice, GenericXlibVideoDevice);

  public:
    ShmXlibVideoDevice(const PString & s)
      : GenericXlibVideoDevice(s) { }

  protected:
    // overrides from GenericXlibVideoDevice
    void OpenRender(Visual * xvisual);
    void CloseRender();
    void DisplayImage(XImage *xi);

    // new functions
    void CreateShmXImage (XImage **image,       XShmSegmentInfo *xsi, Visual * visual);
    void DestroyShmXImage(XImage **SourceImage, XShmSegmentInfo *xsi);

    // member variables
    XShmSegmentInfo shminfo0;
    XShmSegmentInfo shminfo1;
};


class XlibVideoDevice : public GenericXlibVideoDevice
{
  PCLASSINFO(XlibVideoDevice, GenericXlibVideoDevice);

  public:
    XlibVideoDevice(const PString & s)
      : GenericXlibVideoDevice(s) { }

  protected:
    // overrides from GenericXlibVideoDevice
    void OpenRender(Visual * xvisual);
    void CloseRender();
    void DisplayImage(XImage *xi);

    // new functions
    void CreateXImage (XImage **image, BYTE ** pixels, Visual * xvisual);
    void DestroyXImage(XImage **SourceImage);
};

#endif
